import 'package:flutter/material.dart';

import '../provider/data.dart';
import '../utils/board_animate.dart';
import '../utils/board_utils.dart';

class Cell {
  int value;
  bool isFixed;
  int conflictsCount = 0;

  bool isAnimated = false;

  Cell(
    @required this.value,
    @required this.isFixed,
  );

  /*
  * Build widget for board cell, with interactions
  */
  Container widget(Data myProvider, int row, int col) {
    String imageAsset = this.getImageAssetName(myProvider);

    return Container(
      decoration: BoxDecoration(
        color: this.getBackgroundColor(myProvider),
        border: this.getCellBorders(myProvider, row, col),
      ),
      child: GestureDetector(
        child: AnimatedSwitcher(
          duration: const Duration(milliseconds: 100),
          transitionBuilder: (Widget child, Animation<double> animation) {
            return ScaleTransition(child: child, scale: animation);
          },
          child: Image(
            image: AssetImage(imageAsset),
            fit: BoxFit.fill,
            key: ValueKey<int>(imageAsset.hashCode),
          ),
        ),
        onTap: () {
          if (col != null && row != null) {
            if (col != myProvider.currentCellCol || row != myProvider.currentCellRow) {
              myProvider.selectCell(col, row);
            } else {
              myProvider.selectCell(null, null);
            }
          }
        },
      )
    );
  }

  /*
  * Build widget for select/update value cell, with interactions
  */
  Container widgetUpdateValue(Data myProvider) {
    if (this.value < 0) {
      return Container();
    }

    String imageAsset = this.getImageAssetName(myProvider);
    Color backgroundColor = Colors.grey[200];

    if (
      myProvider.showConflicts
      && myProvider.currentCellCol != null
      && myProvider.currentCellRow != null
    ) {
      List cells = myProvider.cells;
      int blockSizeHorizontal = myProvider.blockSizeHorizontal;
      int blockSizeVertical = myProvider.blockSizeVertical;

      if (!BoardUtils.isValueAllowed(cells, blockSizeHorizontal, blockSizeVertical, myProvider.currentCellCol, myProvider.currentCellRow, this.value)) {
        backgroundColor = Colors.pink[100];
      }
    }

    return Container(
      decoration: BoxDecoration(
        color: backgroundColor,
        border: Border.all(
          color: Colors.black,
          width: 2,
        ),
      ),
      child: GestureDetector(
        child: Image(
          image: AssetImage(imageAsset),
          fit: BoxFit.fill
        ),
        onTap: () {
          if (myProvider.currentCellCol != null && myProvider.currentCellRow != null) {
            myProvider.updateCellValue(myProvider.currentCellCol, myProvider.currentCellRow, this.value);
          }
          myProvider.selectCell(null, null);
          if (BoardUtils.checkBoardIsSolved(myProvider)) {
            BoardAnimate.startAnimation(myProvider, 'win');
          }
        },
      )
    );
  }

  /*
  * Compute image asset name, from skin and cell value/state
  */
  String getImageAssetName(Data myProvider) {
    String imageAsset = 'assets/skins/empty.png';
    if (this.value > 0) {
      int cellValue = myProvider.getTranslatedValueForDisplay(this.value);
      imageAsset = 'assets/skins/' + myProvider.skin + '_' + cellValue.toString() + '.png';
    }

    return imageAsset;
  }

  // Compute cell background color, from cell state
  Color getBackgroundColor(Data myProvider) {
    Color editableCellColor = Colors.grey[100];
    Color editableCellColorConflict = Colors.pink[100];
    Color fixedCellColor = Colors.grey[300];
    Color fixedCellColorConflict = Colors.pink[200];
    Color editableSelectedValueColor = Colors.green[100];
    Color fixedSelectedValueColor = Colors.green[300];
    Color editableAnimated = Colors.green[200];
    Color fixedAnimated = Colors.green[300];

    Color backgroundColor = editableCellColor;

    if (this.isFixed) {
      backgroundColor = fixedCellColor;
    }

    if (myProvider.showConflicts && (this.conflictsCount != 0)) {
      if (this.isFixed) {
        backgroundColor = fixedCellColorConflict;
      } else {
        backgroundColor = editableCellColorConflict;
      }
    }

    if (myProvider.showConflicts && (this.value == myProvider.currentCellValue)) {
      if (this.isFixed) {
        backgroundColor = fixedSelectedValueColor;
      } else {
        backgroundColor = editableSelectedValueColor;
      }
    }

    if (this.isAnimated) {
      if (this.isFixed) {
        backgroundColor = fixedAnimated;
      } else {
        backgroundColor = editableAnimated;
      }
    }

    return backgroundColor;
  }

  // Compute cell borders, from board size and cell state
  Border getCellBorders(Data myProvider, int row, int col) {
    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    Color cellBorderDarkColor = Colors.black;
    Color cellBorderLightColor = Colors.grey;
    Color cellBorderSelectedColor = Colors.red;

    Color cellBorderColor = cellBorderSelectedColor;
    double cellBorderWidth = 4;

    // Reduce cell border width on big boards
    int boardSize = blockSizeVertical * blockSizeHorizontal;
    if (boardSize > 8) {
        cellBorderWidth = 2;
        if (boardSize > 10) {
            cellBorderWidth = 1;
        }
    }

    if (!myProvider.gameIsRunning) {
      cellBorderColor = Colors.green[700];
    }

    Border borders = Border.all(
      color: cellBorderColor,
      width: cellBorderWidth,
    );

    // Update cell borders if not currently selected cell
    if (col != myProvider.currentCellCol || row != myProvider.currentCellRow) {
      borders = Border(
        top: BorderSide(width: cellBorderWidth, color: ((row % blockSizeVertical) == 0) ? cellBorderDarkColor : cellBorderLightColor),
        left: BorderSide(width: cellBorderWidth, color: ((col % blockSizeHorizontal) == 0) ? cellBorderDarkColor : cellBorderLightColor),
        right: BorderSide(width: cellBorderWidth, color: (((col + 1) % blockSizeHorizontal) == 0) ? cellBorderDarkColor : cellBorderLightColor),
        bottom: BorderSide(width: cellBorderWidth, color: (((row + 1) % blockSizeVertical) == 0) ? cellBorderDarkColor : cellBorderLightColor),
      );
    }

    return borders;
  }
}
