import 'package:flutter/foundation.dart';
import 'package:shared_preferences/shared_preferences.dart';

class Data extends ChangeNotifier {

  // Configuration available values
  List _availableDifficultyLevels = ['easy', 'medium', 'hard', 'nightmare'];
  List _availableSizes = ['2x2', '3x2', '3x3', '4x4'];
  List _availableSkins = ['default', 'food', 'nature', 'monsters'];
  List _shufflableSkins = ['food', 'nature', 'monsters'];

  List get availableDifficultyLevels => _availableDifficultyLevels;
  List get availableSizes => _availableSizes;
  List get availableSkins => _availableSkins;

  // Application default configuration
  String _level = null;
  String _levelDefault = 'medium';
  String _size = null;
  String _sizeDefault = '3x3';
  String _skin = null;
  String _skinDefault = 'default';

  // Game data
  bool _assetsPreloaded = false;
  bool _gameIsRunning = false;
  bool _animationInProgress = false;
  int _blockSizeVertical = null;
  int _blockSizeHorizontal = null;
  List _cells = [];
  List _cellsSolved = [];
  List _shuffledCellValues = [];
  int _currentCellCol = null;
  int _currentCellRow = null;
  int _currentCellValue = null;
  bool _showConflicts = false;
  int _givenTipsCount = 0;

  String get level => _level;
  void updateLevel(String level) {
    _level = level;
    notifyListeners();
  }

  String get size => _size;
  int get blockSizeVertical => _blockSizeVertical;
  int get blockSizeHorizontal => _blockSizeHorizontal;
  void updateSize(String size) {
    _size = size;
    _blockSizeHorizontal = int.parse(_size.split('x')[0]);
    _blockSizeVertical = int.parse(_size.split('x')[1]);
    notifyListeners();
  }

  String get skin => _skin;
  void updateSkin(String skin) {
    _skin = skin;
    notifyListeners();
  }

  getParameterValue(String parameterCode) {
    switch(parameterCode) {
      case 'difficulty': { return _level; }
      break;
      case 'size': { return _size; }
      break;
      case 'skin': { return _skin; }
      break;
    }
  }

  List getParameterAvailableValues(String parameterCode) {
    switch(parameterCode) {
      case 'difficulty': { return _availableDifficultyLevels; }
      break;
      case 'size': { return _availableSizes; }
      break;
      case 'skin': { return _availableSkins; }
      break;
    }
  }

  setParameterValue(String parameterCode, String parameterValue) async {
    switch(parameterCode) {
      case 'difficulty': { updateLevel(parameterValue); }
      break;
      case 'size': { updateSize(parameterValue); }
      break;
      case 'skin': { updateSkin(parameterValue); }
      break;
    }
    final prefs = await SharedPreferences.getInstance();
    prefs.setString(parameterCode, parameterValue);
  }

  void initParametersValues() async {
    final prefs = await SharedPreferences.getInstance();
    setParameterValue('difficulty', prefs.getString('difficulty') ?? _levelDefault);
    setParameterValue('size', prefs.getString('size') ?? _sizeDefault);
    setParameterValue('skin', prefs.getString('skin') ?? _skinDefault);
  }

  bool get gameIsRunning => _gameIsRunning;
  void updateGameIsRunning(bool gameIsRunning) {
    _gameIsRunning = gameIsRunning;
    notifyListeners();
  }

  bool get assetsPreloaded => _assetsPreloaded;
  void updateAssetsPreloaded(bool assetsPreloaded) {
    _assetsPreloaded = assetsPreloaded;
  }

  List get cells => _cells;
  void updateCells(List cells) {
    _cells = cells;
    notifyListeners();
  }

  List get cellsSolved => _cellsSolved;
  void updateCellsSolved(List cells) {
    _cellsSolved = cells;
  }

  void shuffleCellValues() {
    int maxCellValue = 16;
    List values = new List<int>.generate(maxCellValue, (i) => i + 1);

    if (_shufflableSkins.contains(_skin)) {
      values.shuffle();
      print('Shuffled tiles values: ' + values.toString());
    }

    _shuffledCellValues = values;
  }
  int getTranslatedValueForDisplay(int originalValue) {
    return _shuffledCellValues[originalValue - 1];
  }

  int get currentCellCol => _currentCellCol;
  set updateCurrentCellCol(int currentCellCol) {
    _currentCellCol = currentCellCol;
    notifyListeners();
  }

  int get currentCellRow => _currentCellRow;
  set updateCurrentCellRow(int currentCellRow) {
    _currentCellRow = currentCellRow;
    notifyListeners();
  }

  int get currentCellValue => _currentCellValue;
  set updateCurrentCellValue(int currentCellValue) {
    _currentCellValue = currentCellValue;
    notifyListeners();
  }

  int get givenTipsCount => _givenTipsCount;
  increaseGivenTipsCount() {
    _givenTipsCount = _givenTipsCount + 1;
    notifyListeners();
  }
  resetGivenTipsCount() {
    _givenTipsCount = 0;
    notifyListeners();
  }

  selectCell(int col, int row) {
    _currentCellCol = col;
    _currentCellRow = row;
    _currentCellValue = null;
    if (row != null && col != null) {
      if (_cells[row][col].value != 0) {
        _currentCellValue = _cells[row][col].value;
      }
    }
    notifyListeners();
  }

  updateCellValue(int col, int row, int value) {
    if (!_cells[row][col].isFixed) {
      _cells[row][col].value = value;
    }
    notifyListeners();
  }

  bool get showConflicts => _showConflicts;
  void updateShowConflicts(bool showConflicts) {
    _showConflicts = showConflicts;
    notifyListeners();
  }
  void toggleShowConflicts() {
    updateShowConflicts(!showConflicts);
  }

  bool get animationInProgress => _animationInProgress;
  void updateAnimationInProgress(bool animationInProgress) {
    _animationInProgress = animationInProgress;
    notifyListeners();
  }

  void setAnimatedBackground(List animatedCellsPattern) {
    int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (var row = 0; row < boardSideLength; row++) {
      for (var col = 0; col < boardSideLength; col++) {
        _cells[row][col].isAnimated = animatedCellsPattern[row][col];
      }
    }
    notifyListeners();
  }
  void resetAnimatedBackground() {
    int boardSideLength = _blockSizeHorizontal * _blockSizeVertical;
    for (var row = 0; row < boardSideLength; row++) {
      for (var col = 0; col < boardSideLength; col++) {
        _cells[row][col].isAnimated = false;
      }
    }
  }
}
