import 'dart:math';

import '../entities/cell.dart';
import '../provider/data.dart';
import '../utils/random_pick_grid.dart';

class BoardUtils {

  static printGrid(List cells, List solvedCells) {
    String stringValues = '0123456789ABCDEFG';
    print('');
    print('-------');
    for (var rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      String row = '';
      String rowSolved = '';
      for (var colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row += stringValues[cells[rowIndex][colIndex].value];
        rowSolved += stringValues[solvedCells[rowIndex][colIndex].value];
      }
      print(row + ' | ' + rowSolved);
    }
    print('-------');
    print('');
  }

  static Future<void> pickGrid(Data myProvider) async {
    String grid;
    RandomPickGrid randomPickGrid;

    randomPickGrid = RandomPickGrid();
    await randomPickGrid.init(myProvider.level, myProvider.size);

    if (randomPickGrid.grid != null) {
      grid = randomPickGrid.grid;
    }

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    if (grid.length == pow(blockSizeHorizontal * blockSizeVertical, 2)) {
      print('Picked grid from template: ' + grid);
      bool isSymetric = (blockSizeHorizontal == blockSizeVertical);
      myProvider.updateCells(BoardUtils.createBoardFromTemplate(grid, isSymetric));
      myProvider.updateCellsSolved(BoardUtils.getSolvedGrid(myProvider));
      myProvider.selectCell(null, null);

      printGrid(myProvider.cells, myProvider.cellsSolved);
    }
  }

  static List createEmptyBoard(int boardSize) {
    List cells = [];
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        row.add(Cell(0, false));
      }
      cells.add(row);
    }

    return cells;
  }

  static List copyBoard(List cells) {
    List copiedGrid = [];
    for (var rowIndex = 0; rowIndex < cells.length; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < cells[rowIndex].length; colIndex++) {
        row.add(Cell(cells[rowIndex][colIndex].value, false));
      }
      copiedGrid.add(row);
    }

    return copiedGrid;
  }

  static List createBoardFromTemplate(String grid, bool isSymetric) {
    List cells = [];
    int boardSize = int.parse(pow(grid.length, 1/2).toStringAsFixed(0));

    String stringValues = '0123456789ABCDEFG';

    int index = 0;
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      List row = [];
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        String stringValue = grid[index++];
        int value = stringValues.indexOf(stringValue);
        row.add(Cell(value, (value != 0)));
      }
      cells.add(row);
    }

    List<String> allowedFlip = ['none', 'horizontal', 'vertical'];
    List<String> allowedRotate = ['none', 'left', 'right'];

    // Forbid rotation if blocks are not symetric
    if (!isSymetric) {
      allowedRotate = ['none'];
    }

    var rand = new Random();
    String flip = allowedFlip[rand.nextInt(allowedFlip.length)];
    String rotate = allowedRotate[rand.nextInt(allowedRotate.length)];

    print('flip board: ' + flip);
    print('rotate board: ' + rotate);

    switch(flip) {
      case 'horizontal': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[boardSize - rowIndex - 1][colIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'vertical': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[rowIndex][boardSize - colIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    switch(rotate) {
      case 'left': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[colIndex][boardSize - rowIndex - 1].value;
          }
        }
        cells = transformedBoard;
      }
      break;
      case 'right': {
        List transformedBoard = createEmptyBoard(boardSize);
        for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
          for (var colIndex = 0; colIndex < boardSize; colIndex++) {
            transformedBoard[rowIndex][colIndex].value = cells[boardSize - colIndex - 1][rowIndex].value;
          }
        }
        cells = transformedBoard;
      }
      break;
    }

    // Fix cells fixed states
    for (var rowIndex = 0; rowIndex < boardSize; rowIndex++) {
      for (var colIndex = 0; colIndex < boardSize; colIndex++) {
        cells[rowIndex][colIndex].isFixed = (cells[rowIndex][colIndex].value != 0) ? true : false;
      }
    }

    return cells;
  }

  static bool checkBoardIsSolved(Data myProvider) {
    List cells = myProvider.cells;

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    int boardSize = blockSizeHorizontal * blockSizeVertical;

    bool isSolved = true;

    // (re)compute conflicts
    BoardUtils.computeConflictsInBoard(myProvider);

    // check grid is fully completed and does not contain conflict
    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        if (cells[row][col].value == 0 || cells[row][col].conflictsCount != 0) {
          return false;
        }
      }
    }

    print('-> ok sudoku solved!');

    return true;
  }

  static bool isValueAllowed(List cells, int blockSizeHorizontal, int blockSizeVertical, int candidateCol, int candidateRow, int candidateValue) {
    if (candidateValue == 0) {
      return true;
    }

    int boardSize = blockSizeHorizontal * blockSizeVertical;

    // check lines does not contains a value twice
    for (var row = 0; row < boardSize; row++) {
      List values = [];
      for (var col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (row == candidateRow && col == candidateCol) {
          value = candidateValue;
        }
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        return false;
      }
    }

    // check columns does not contains a value twice
    for (var col = 0; col < boardSize; col++) {
      List values = [];
      for (var row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (row == candidateRow && col == candidateCol) {
          value = candidateValue;
        }
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        return false;
      }
    }

    // check blocks does not contains a value twice
    int horizontalBlocksCount = blockSizeVertical;
    int verticalBlocksCount = blockSizeHorizontal;
    for (var blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (var blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        List values = [];

        for (var rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (var colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            int row = (blockRow * blockSizeVertical) + rowInBlock;
            int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (row == candidateRow && col == candidateCol) {
              value = candidateValue;
            }
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          return false;
        }
      }
    }

    return true;
  }

  static void computeConflictsInBoard(Data myProvider) {
    List cells = myProvider.cells;

    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    int boardSize = blockSizeHorizontal * blockSizeVertical;

    // reset conflict states
    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        cells[row][col].conflictsCount = 0;
      }
    }

    // check lines does not contains a value twice
    for (var row = 0; row < boardSize; row++) {
      List values = [];
      for (var col = 0; col < boardSize; col++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('line ' + row.toString() + ' contains duplicates');
        // Add line to cells in conflict
        for (var col = 0; col < boardSize; col++) {
          cells[row][col].conflictsCount++;
        }
      }
    }

    // check columns does not contains a value twice
    for (var col = 0; col < boardSize; col++) {
      List values = [];
      for (var row = 0; row < boardSize; row++) {
        int value = cells[row][col].value;
        if (value != 0) {
          values.add(value);
        }
      }
      List distinctValues = values.toSet().toList();
      if (values.length != distinctValues.length) {
        print('column ' + col.toString() + ' contains duplicates');
        // Add column to cells in conflict
        for (var row = 0; row < boardSize; row++) {
          cells[row][col].conflictsCount++;
        }
      }
    }

    // check blocks does not contains a value twice
    int horizontalBlocksCount = blockSizeVertical;
    int verticalBlocksCount = blockSizeHorizontal;
    for (var blockRow = 0; blockRow < verticalBlocksCount; blockRow++) {
      for (var blockCol = 0; blockCol < horizontalBlocksCount; blockCol++) {
        List values = [];

        for (var rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
          for (var colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
            int row = (blockRow * blockSizeVertical) + rowInBlock;
            int col = (blockCol * blockSizeHorizontal) + colInBlock;
            int value = cells[row][col].value;
            if (value != 0) {
              values.add(value);
            }
          }
        }

        List distinctValues = values.toSet().toList();
        if (values.length != distinctValues.length) {
          print('block [' + blockCol.toString() + ',' + blockRow.toString() + '] contains duplicates');
          // Add blocks to cells in conflict
          for (var rowInBlock = 0; rowInBlock < blockSizeVertical; rowInBlock++) {
            for (var colInBlock = 0; colInBlock < blockSizeHorizontal; colInBlock++) {
              int row = (blockRow * blockSizeVertical) + rowInBlock;
              int col = (blockCol * blockSizeHorizontal) + colInBlock;
              cells[row][col].conflictsCount++;
            }
          }
        }
      }
    }
  }

  static List getCellsWithWrongValue(List cells, List cellsSolved, int blockSizeHorizontal, int blockSizeVertical) {
    List cellsWithWrongValue = [];
    int boardSize = blockSizeHorizontal * blockSizeVertical;

    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        if (cells[row][col].value != 0 && cells[row][col].value != cellsSolved[row][col].value) {
          cellsWithWrongValue.add([col, row]);
        }
      }
    }

    return cellsWithWrongValue;
  }

  static List getConflictingCells(List cells, int blockSizeHorizontal, int blockSizeVertical) {
    List conflictingCells = [];
    int boardSize = blockSizeHorizontal * blockSizeVertical;

    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        if (!cells[row][col].isFixed && cells[row][col].value != 0) {
          if (cells[row][col].conflictsCount != 0 && !BoardUtils.isValueAllowed(cells, blockSizeHorizontal, blockSizeVertical, col, row, cells[row][col].value)) {
            conflictingCells.add([col, row]);
          }
        }
      }
    }

    return conflictingCells;
  }

  static List getCellsWithUniqueAvailableValue(List cells, int blockSizeHorizontal, int blockSizeVertical) {
    List candidateCells = [];
    int boardSize = blockSizeHorizontal * blockSizeVertical;

    for (var row = 0; row < boardSize; row++) {
      for (var col = 0; col < boardSize; col++) {
        if (cells[row][col].value == 0) {
          int allowedValuesCount = 0;
          int candidateValue = 0;
          for (var value = 1; value <= boardSize; value++) {
            if (BoardUtils.isValueAllowed(cells, blockSizeHorizontal, blockSizeVertical, col, row, value)) {
              candidateValue = value;
              allowedValuesCount++;
            }
          }
          if (allowedValuesCount == 1) {
            candidateCells.add([col, row, candidateValue]);
          }
        }
      }
    }

    return candidateCells;
  }

  static List getSolvedGrid(Data myProvider) {
    List cells = copyBoard(myProvider.cells);
    int blockSizeHorizontal = myProvider.blockSizeHorizontal;
    int blockSizeVertical = myProvider.blockSizeVertical;

    do {
      List cellsWithUniqueAvailableValue = BoardUtils.getCellsWithUniqueAvailableValue(cells, blockSizeHorizontal, blockSizeVertical);
      if (cellsWithUniqueAvailableValue.length == 0) {
        break;
      }
      for (var i = 0; i < cellsWithUniqueAvailableValue.length; i++) {
        int col = cellsWithUniqueAvailableValue[i][0];
        int row = cellsWithUniqueAvailableValue[i][1];
        int value = cellsWithUniqueAvailableValue[i][2];
        cells[row][col].value = value;
      }
    } while (true);

    return cells;
  }

}
